@@ -33,7 +33,11 @@ struct imp_dbh_st {
     int  odbc_ignore_named_placeholders;	/* flag to ignore named parameters */
     int  odbc_default_bind_type;	/* flag to set default binding type (experimental) */
     int  odbc_sqldescribeparam_supported; /* flag to see if SQLDescribeParam is supported */
-    int  odbc_sqlmoreresults_supported; /* flag to see if SQLDescribeParam is supported */
+    int  odbc_sqlmoreresults_supported; /* flag to see if SQLMoreResults is supported */
+    int  odbc_async_exec; /* flag to set asynchronous execution */
+    SQLUINTEGER odbc_async_type; /* flag to store the type of asynchronous
+                                  * execution the driver supports */
+    SV *odbc_err_handler; /* contains the error handler coderef */
 };
 
 /* Define sth implementor data structure */
@@ -44,7 +48,7 @@ struct imp_sth_st {
     HDBC       hdbc;		/* copy for speed	*/
     HSTMT      hstmt;
 
-	int        moreResults;	/* are there more results to fetch?	*/
+    int        moreResults;	/* are there more results to fetch?	*/
     int        done_desc;	/* have we described this sth yet ?	*/
 
     /* Input Details	*/
